/*
 * proc_service implementation for live processes
 */

#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <sys/ptrace.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/procfs.h>

#include "proc_service.h"
#include "proc_service_impl.h"

/*
 * Copies size bytes from the data segment of the target process to the
 * controlling process.
 */
static ps_err_e ptrace_pdread(struct ps_prochandle *ph, psaddr_t addr,
			      void *buf, size_t size)
{
	char *ptr = buf;
	long res;
	long i, words = (size + (sizeof(long) - 1))/sizeof(long);

	for (i = 0; i < words; i++) {
		res = ptrace(PTRACE_PEEKDATA, ph->pid, addr, 0);
		if (errno) {
			ps_debug("ptrace failed: %s\n", strerror(errno));
			return PS_ERR;
		}

		if (size >= sizeof(long)) {
			*(long *)ptr = res;
			ptr += sizeof(long);
			size -= sizeof(long);
		} else {
			char *tmp = (char *)&res;
			for (; size > 0; size--)
				*ptr++ = *tmp++;
		}

		addr += sizeof(long);
	}

	return PS_OK;
}

static ps_err_e ptrace_lgetregs(struct ps_prochandle *ph, lwpid_t pid,
				prgregset_t regset)
{
	long res;

	res = ptrace(PTRACE_GETREGS, ph->pid, NULL, regset);
	if (res == -1) {
		ps_debug("ptrace failed: %s\n", strerror(errno));
		return PS_ERR;
	}

	return PS_OK;
}

static void ptrace_release(struct ps_prochandle *ph)
{
	long res;

	res = ptrace(PTRACE_DETACH, ph->pid, NULL, NULL);
	if (res == -1)
		ps_debug("ptrace failed: %s\n", strerror(errno));
}

static struct ps_prochandle_ops ps_ptrace_ops = {
	.release	= ptrace_release,
	.pdread		= ptrace_pdread,
	.lgetregs	= ptrace_lgetregs,
};

struct ps_prochandle* ptrace_attach(pid_t pid)
{
	int res;
	int status;
	struct ps_prochandle* ph;
	char *filename;

	ph = calloc(1, sizeof(*ph));
	if (!ph) {
		ps_debug("Couldn't allocate memory\n");
		return NULL;
	}

	res = ptrace(PTRACE_ATTACH, pid, 0, 0);
	if (res == -1) {
		ps_debug("ptrace failed: %s\n", strerror(errno));
		free(ph);
		return NULL;
	}

	res = waitpid(pid, &status, WUNTRACED);

	/*
	 * If we failed due to an ECHILD, then retry with the __WCLONE
	 * flag.  Note we loop as the the PID we get back may not be
	 * one we care about.
	 */
	if (res == -1) {
		if (errno != ECHILD) {
			free(ph);
			return NULL;
		}
		while (1) {
			res = waitpid (-1, &status, (__WCLONE));
			if (res == -1) {
				ps_debug("waitpid failed: %s\n",
					 strerror(errno));
				free(ph);
				return NULL;
			} else if (res == pid)
				break;
		}
	}

	if (WIFEXITED(status)) {
		ps_debug("process exited\n");
		ps_release(ph);
		free(ph);
		return NULL;
	}

	ph->pid = pid;
	ph->ops = &ps_ptrace_ops;
	return ph;
}
