/*
 * symbols.c - A symbol table implementation from pstack
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * FIXME: This code can only handle native ELF!
 */

#include <fcntl.h>
#include <link.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <unistd.h>

#include "proc_service.h"
#include "symbols.h"

struct Elf_Ehdr {
	union {
		Elf32_Ehdr e32;
		Elf64_Ehdr e64;
	} e_un;
	int is_64bit;
};

struct Elf_Shdr {
	union {
		Elf32_Shdr s32;
		Elf64_Shdr s64;
	} s_un;
	int is_64bit;
};

struct Elf_Sym {
	union {
		Elf32_Sym s32;
		Elf64_Sym s64;
	} s_un;
	int is_64bit;
};

#define Elf_Ehdr(x,name) ((x)->is_64bit ? \
			  (x)->e_un.e64.name :	\
			  (x)->e_un.e32.name)

#define Elf_Shdr(x,name) ((x)->is_64bit ? \
			  (x)->s_un.s64.name :	\
			  (x)->s_un.s32.name)

#define Elf_Sym(x,name) ((x)->is_64bit ? \
			 (x)->s_un.s64.name :	\
			 (x)->s_un.s32.name)

struct Symbols {
	struct Symbols *next;
	char *name;
	struct Elf_Sym *symbols;
	int nsyms;
	char *strings;
	int strslen, noffsets;
	ElfW(Addr) baseAddr;
	ElfW(Dyn) *dynamic;
	int ndyns;
	int is_64bit;
};

static Symbols_t allSyms;

static Symbols_t newSyms(const char *name)
{
	Symbols_t syms = (Symbols_t) calloc(1, sizeof(struct Symbols));

	if (!syms)
		ps_abort("Out of memory");
	syms->next = allSyms;
	allSyms = syms;
	syms->name = strdup(name);
	return syms;
}

static void deleteSyms(Symbols_t syms)
{
	Symbols_t s2;

	if (syms == allSyms)
		allSyms = syms->next;
	else {
		for (s2 = allSyms; s2 && s2->next != syms; s2 = s2->next);
		if (s2)
			s2->next = syms->next;
	}
	if (syms->symbols)
		free(syms->symbols);
	if (syms->strings)
		free(syms->strings);
	if (syms->dynamic)
		free(syms->dynamic);
	if (syms->name)
		free(syms->name);
	free(syms);
}

const Elf32_Sym *lookupSym32InTable(const char *name, Symbols_t syms)
{
	struct Elf_Sym *sym;
	int i;

	for (i = 0, sym = syms->symbols; i < syms->nsyms; i++, sym++) {
		if (!strcmp(name, &syms->strings[Elf_Sym(sym, st_name)]))
			return &sym->s_un.s32;
	}

	return 0;
}

const Elf64_Sym *lookupSym64InTable(const char *name, Symbols_t syms)
{
	struct Elf_Sym *sym;
	int i;

	for (i = 0, sym = syms->symbols; i < syms->nsyms; i++, sym++) {
		if (!strcmp(name, &syms->strings[Elf_Sym(sym, st_name)]))
			return &sym->s_un.s64;
	}

	return 0;
}

void findCodeAddress(ElfW(Addr) addr, ElfW(Sym) ** ans, Symbols_t * symtab)
{
	struct Elf_Sym *sym;
	Symbols_t tab;
	int i;

	for (tab = allSyms, *ans = 0, *symtab = 0; tab; tab = tab->next) {
		if (addr < tab->baseAddr)
			continue;
		for (sym = tab->symbols, i = 0; i < tab->nsyms; i++, sym++) {
			if (Elf_Sym(sym, st_value) <= addr
			    && Elf_Sym(sym, st_shndx) != SHN_UNDEF
			    && Elf_Sym(sym, st_shndx) < tab->noffsets
			    && ELF32_ST_TYPE(Elf_Sym(sym, st_info)) == STT_FUNC
			    && (!*ans || (*ans)->st_value < Elf_Sym(sym, st_value)))
				*ans = sym, *symtab = tab;
		}
	}
}

const Elf64_Addr findLocalSym(const char *name, Symbols_t syms)
{
	union {
		Elf32_Sym *s32;
		Elf64_Sym *s64;
	} sym_un;

	if (syms->is_64bit)
		sym_un.s64 = lookupSym64InTable(name, syms);
	else
		sym_un.s32 = lookupSym32InTable(name, syms);

	if (syms->is_64bit)
		return (!sym_un.s64 || sym_un.s64->st_shndx == SHN_UNDEF ||
		sym_un.s64->st_shndx >= syms->noffsets) ?
			0 :
			sym_un.s64->st_value;
	else
		return (!sym_un.s32 || sym_un.s32->st_shndx == SHN_UNDEF ||
		sym_un.s32->st_shndx >= syms->noffsets) ?
			0 :
			sym_un.s32->st_value;
}

static void verify_ident(struct Elf_Ehdr *hdr)
{
	if (memcmp(&hdr->e_un.e32.e_ident[EI_MAG0], ELFMAG, SELFMAG))
		ps_abort("Bad magic number.");
	if ((hdr->e_un.e32.e_ident[EI_CLASS] != ELFCLASS32) &&
	    (hdr->e_un.e32.e_ident[EI_CLASS] != ELFCLASS64))
		ps_abort("only 32 bit and 64 bit objects supported.");

	hdr->is_64bit = (hdr->e_un.e32.e_ident[EI_CLASS] == ELFCLASS64);
	ps_debug("This is a %dbit process\n", hdr->is_64bit ? 64 : 32);

	if (hdr->e_un.e32.e_ident[EI_DATA] != ELFDATA2LSB)
		ps_abort("big endian object files not supported.");
	if (hdr->e_un.e32.e_ident[EI_VERSION] != EV_CURRENT ||
	    hdr->e_un.e32.e_version != EV_CURRENT)
		ps_abort("Unsupported ELF format version.");
	if ((hdr->e_un.e32.e_machine != EM_386) &&
	    (hdr->e_un.e32.e_machine != EM_X86_64))
		ps_abort("Not an IA32 or AMD64 executable.");
}

static int find_symtabs(struct Elf_Ehdr *hdr, int fd, Symbols_t syms)
{
	int i, idx, spot;
	struct Elf_Shdr shdr;

	spot = Elf_Ehdr(hdr,e_shoff);
	if (lseek(fd, spot, SEEK_SET) != spot)
		ps_abort("seek failed.");

	memset(&shdr, 0, sizeof(shdr));

	shdr.is_64bit = hdr->is_64bit;
	syms->is_64bit = hdr->is_64bit;

	syms->noffsets = Elf_Ehdr(hdr,e_shnum);

	for (idx = 0; idx < Elf_Ehdr(hdr,e_shnum); idx++) {
		if (read(fd, &shdr, Elf_Ehdr(hdr,e_shentsize))
		    != Elf_Ehdr(hdr,e_shentsize))
			ps_abort("premature eof.");
		spot += Elf_Ehdr(hdr,e_shentsize);

		ps_debug("sh_type = %x\n", Elf_Shdr(&shdr,sh_type));

		switch (Elf_Shdr(&shdr,sh_type)) {
		case SHT_SYMTAB:
			syms->nsyms = Elf_Shdr(&shdr,sh_size) /
				(shdr.is_64bit ? sizeof(Elf64_Sym) :
				 sizeof(Elf32_Sym));

			syms->symbols = malloc(Elf_Shdr(&shdr,sh_size));
			if (!syms->symbols)
				ps_abort("Could not allocate symbol table.");

			if (lseek(fd, Elf_Shdr(&shdr,sh_offset), SEEK_SET) !=
			    Elf_Shdr(&shdr,sh_offset)
			    || read(fd, syms->symbols,
				    Elf_Shdr(&shdr,sh_size))
			    != Elf_Shdr(&shdr,sh_size))
				ps_abort("Could not read symbol table.");

			i = Elf_Ehdr(hdr,e_shoff) + Elf_Shdr(&shdr,sh_link) * Elf_Ehdr(hdr,e_shentsize);
			if (lseek(fd, i, SEEK_SET) != i)
				ps_abort("Could not seek and find.");
			if (read(fd, &shdr, Elf_Ehdr(hdr,e_shentsize)) !=
			    Elf_Ehdr(hdr,e_shentsize))
				ps_abort("Could not read string table section header.");
			if (!(syms->strings = malloc(Elf_Shdr(&shdr,sh_size))))
				ps_abort("Could not allocate string table.");
			if (lseek(fd, Elf_Shdr(&shdr,sh_offset), SEEK_SET) !=
			    Elf_Shdr(&shdr,sh_offset)
			    || read(fd, syms->strings,
				    Elf_Shdr(&shdr,sh_size)) != Elf_Shdr(&shdr,sh_size))
				ps_abort("Could not read string table.");
			lseek(fd, spot, SEEK_SET);
			break;
		case SHT_DYNAMIC:
			syms->ndyns = Elf_Shdr(&shdr,sh_size) /
				(shdr.is_64bit ? sizeof(Elf64_Dyn) :
				 sizeof(Elf32_Dyn));
			if (!
			    (syms->dynamic = malloc(Elf_Shdr(&shdr,sh_size))))
				ps_abort("Out of memory.");
			if (lseek(fd, Elf_Shdr(&shdr,sh_offset), SEEK_SET) !=
			    Elf_Shdr(&shdr,sh_offset)
			    || read(fd, syms->dynamic,
				    Elf_Shdr(&shdr,sh_size)) != Elf_Shdr(&shdr,sh_size))
				ps_abort("Could not read dynamic table.");
			lseek(fd, spot, SEEK_SET);
			break;
		}
	}

	ps_debug("Loaded %d symbols (%d dynamic)\n", syms->nsyms + syms->ndyns,
		 syms->ndyns);

	return (syms->nsyms > 0);
}

static void resolveSymbols(Symbols_t syms, ElfW(Addr) baseaddr)
{
	ElfW(Sym) *sym;
	int i;

	syms->baseAddr = baseaddr;

	for (i = 0, sym = syms->symbols; i < syms->nsyms; i++, sym++) {
		if (sym->st_shndx && sym->st_shndx < syms->noffsets) {
			sym->st_value += baseaddr;
		}
	}
}

Symbols_t loadSyms(const char *fname, unsigned long baseaddr)
{
	struct Elf_Ehdr hdr;
	int fd;
	Symbols_t syms;

	syms = newSyms(fname);
	if ((fd = open(fname, O_RDONLY)) < 0) {
		fprintf(stderr, "'%s': ", fname);
		perror("opening object file");
		ps_abort("Could not open object file.");
	}
	read(fd, &hdr, sizeof(hdr));
	verify_ident(&hdr);
	if (!find_symtabs(&hdr, fd, syms)) {
		deleteSyms(syms);
		syms = 0;
	} else
		resolveSymbols(syms, baseaddr);
	close(fd);

	return syms;
}

static void resetData(void)
{
	Symbols_t syms, ns;

	for (syms = allSyms; syms; syms = ns) {
		ns = syms->next;
		deleteSyms(syms);
	}
}

int ps_symbols_init(void)
{
	return 0;
}

void ps_symbols_exit(void)
{
	resetData();
}
